#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <dxr2.h>

#include <plugin/decoder.h>
#include <plugin/buf.h>

static struct {
	pthread_t loop;
	pthread_cond_t playing;
	pthread_mutex_t mutex;
	int end_loop;
	int fd_dxr2;
	int current_audio_mode;
	int current_audio_stream;
	buf_t *buf;
	int status;
} decoder;

static plugin_codec_t codec_dxr2 = {
	DecoderInit,
	NULL,		// exit
        NULL,		// open
        NULL,		// close
	DecoderBufInit,
	DecoderBufExit,
	DecoderBufPut,	// read
	NULL,		// write
	DecoderSetStatus,
	DecoderSetAttributes,
};


/**
 *
 */

plugin_codec_t *codec_init ()
{
	return &codec_dxr2;
}


/**
 *
 */

static void *_DecoderLoop (void *foo) {
	#define BUFSIZE 0x8000
  
	char *buffer=malloc(BUFSIZE);
	int index=0, size;
	buf_entry_t buf_entry;

	ioctl(decoder.fd_dxr2,DXR2_IOC_PLAY,NULL);
	decoder.status=DXR2_PLAYMODE_PLAY;

  // Wait for buffer to fill up
	while((decoder.buf->ctrl.write>=decoder.buf->ctrl.read ?
		decoder.buf->ctrl.write-decoder.buf->ctrl.read :
		decoder.buf->ctrl.write-decoder.buf->ctrl.read+BUF_DEPTH)<BUF_DEPTH/2)
		usleep(10000);
  

	while (!decoder.end_loop) {
		if (decoder.status == DXR2_PLAYMODE_PAUSED) {
			ioctl(decoder.fd_dxr2,DXR2_IOC_PAUSE,NULL);
			pthread_mutex_lock(&(decoder.mutex));
			pthread_cond_wait (&(decoder.playing), &(decoder.mutex));
			pthread_mutex_unlock(&(decoder.mutex));

			ioctl(decoder.fd_dxr2,DXR2_IOC_PLAY,NULL);
			decoder.status=DXR2_PLAYMODE_PLAY;
		}

		if (!DecoderBufGet (decoder.buf, BUF_ANY, &buf_entry)) {
			if (index+buf_entry.mem_len>BUFSIZE) {
				pthread_mutex_lock(&(decoder.mutex));
				size = index;
				while (size>0)
					size-=write (decoder.fd_dxr2, buffer+index-size, size);
				index=0; 
				pthread_mutex_unlock(&(decoder.mutex));
			}

			memcpy (&buffer[index],buf_entry.mem,buf_entry.mem_len);
			index += buf_entry.mem_len;
			free (buf_entry.mem);
		} else {
			pthread_cond_wait (&(decoder.buf->ctrl.cond_new_data),
				&(decoder.buf->ctrl.mutex));
		}
	}
	fprintf(stderr,"Decoder Stopped\n");

	return NULL;
}



static int init_registers() {
	dxr2_threeArg_t buf3;

	buf3.arg1 = DXR2_ASPECTRATIO_4_3;
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_SOURCE_ASPECT_RATIO, &buf3))
		return -1;

	buf3.arg1 = DXR2_MACROVISION_OFF;
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_TV_MACROVISION_MODE, &buf3))
		return -1;

	buf3.arg1 = DXR2_INTERLACED_ON;
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_TV_INTERLACED_MODE, &buf3))
		return -1;

	buf3.arg1 = DXR2_75IRE_OFF;
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_TV_75IRE_MODE, &buf3))
		return -1;

	buf3.arg1 = DXR2_IEC958_ENCODED;
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_IEC958_OUTPUT_MODE, &buf3))
		return -1;

	buf3.arg1 = DXR2_AUDIO_MUTE_OFF;
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_AUDIO_MUTE, &buf3))
		return -1;

	buf3.arg1 = DXR2_CSSDECRMODE_OFF;
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_DECRYPTION_MODE, &buf3))
		return -1;

	DecoderSPU (-1);

	return 0;
}


/**
 *
 */

static int install_firmware(char *ucode) {
	int uCodeFD;
	int uCodeSize;
	dxr2_uCode_t* uCode;

	if ((uCodeFD = open(ucode, O_RDONLY)) < 0) {
		perror(ucode);
		return -1;
	}
	uCodeSize = lseek(uCodeFD, 0, SEEK_END);
	if ((uCode = (dxr2_uCode_t*)malloc(uCodeSize + 4)) == NULL) {
		perror("malloc");
		return -1;
	}
	lseek(uCodeFD, 0, SEEK_SET);
	if (read(uCodeFD, uCode+4, uCodeSize) != uCodeSize) {
		perror(ucode);
		return -1;
	}
	close(uCodeFD);
	uCode->uCodeLength = uCodeSize;

  // upload ucode
	if (ioctl(decoder.fd_dxr2, DXR2_IOC_INIT_ZIVADS, uCode)) {
		fprintf(stderr,"uCode upload failed!\n");
		return -1;
	}

	return 0;
}


/**
 *
 */

int DecoderHighlightButton(int button) {
	dxr2_threeArg_t buf3;
	buf3.arg1 = button;
	buf3.arg2 = DXR2_BUTTONACTION_ACTIVATE_SELECTED;
	return ioctl(decoder.fd_dxr2,DXR2_IOC_HIGHLIGHT,&buf3); 
}


/**
 *
 */

int DecoderInit (buf_t *buf, char *output_path) {
	fprintf(stderr,"DecoderInit\n");

  // open device
	if ((decoder.fd_dxr2 = open(output_path, O_WRONLY)) < 0) {
		perror(output_path);
		return -1;
	}  

	if (install_firmware ("dvd1.ux") == -1)
		return -1;
	if (init_registers() == -1)
		return -1;

	decoder.status = DXR2_PLAYMODE_STOPPED;
	pthread_mutex_init(&(decoder.mutex),NULL);
	pthread_cond_init(&(decoder.playing),NULL);

	decoder.buf = buf;
	decoder.current_audio_mode = -1;
	decoder.current_audio_stream = -1;

	return 0;
}


/**
 *
 */

int DecoderSetVideoAttributes (plugin_codec_attr_video_t *attr)
{
	dxr2_threeArg_t buf3;

	// Change of fps can't be done while playing
	if (decoder.status == DXR2_PLAYMODE_PLAY && attr->fps != -1)
		ioctl(decoder.fd_dxr2,DXR2_IOC_STOP,NULL);


	if (attr->mode != -1) {
		switch (attr->mode) {
		case 0:
			buf3.arg1 = DXR2_BITSTREAM_TYPE_MPEG_VOB;
			break;
		case 1:
			buf3.arg1 = DXR2_BITSTREAM_TYPE_CDROM_VCD;
			break;
		case 2:
			buf3.arg1 = DXR2_BITSTREAM_TYPE_CDDA;
			break;
		default:
			return -1;		
		}
		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_BITSTREAM_TYPE, &buf3)) 
			return -1;
	}

	if (attr->fps != -1) {
		switch (attr->fps) {
		case 0:
			buf3.arg1 = DXR2_SRC_VIDEO_FREQ_25;
			break;
		case 1:
			buf3.arg1 = DXR2_SRC_VIDEO_FREQ_30;
			break;
		default:
			return -1;
		}

		buf3.arg2 = 0x2d0;
		buf3.arg3 = 0x1e0;

		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_SOURCE_VIDEO_FORMAT, &buf3))
			return -1;
	}

	if (attr->aspect_mode != -1) {
		switch (attr->aspect_mode) {
		case 0:
			buf3.arg1 = DXR2_ASPECTRATIOMODE_NORMAL;
			break;
		case 1:
			buf3.arg1 = DXR2_ASPECTRATIOMODE_PAN_SCAN;
			break;
		case 2:
			buf3.arg1 = DXR2_ASPECTRATIOMODE_LETTERBOX;
			break;
		default:
			return -1;
		}
		if (ioctl(decoder.fd_dxr2,DXR2_IOC_SET_ASPECT_RATIO_MODE,&buf3))
			return -1;
	}    

	if (attr->aspect_output != -1) {
		switch (attr->aspect_output) {
		case 0:
			buf3.arg1 = DXR2_ASPECTRATIO_4_3;
			break;
		case 3:
			buf3.arg1 = DXR2_ASPECTRATIO_16_9;
			break;
		default:
			return -1;
		}
		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_OUTPUT_ASPECT_RATIO, &buf3))
			return -1;
	}

	if (attr->tv_system != -1) {
		switch (attr->tv_system) {
		case 0:
			buf3.arg1 = DXR2_OUTPUTFORMAT_NTSC;
			break;				
		case 1:
			buf3.arg1 = DXR2_OUTPUTFORMAT_NTSC_60;
			break;
		case 2:
			buf3.arg1 = DXR2_OUTPUTFORMAT_PAL_BDGHI;
			break;
		case 3:
			buf3.arg1 = DXR2_OUTPUTFORMAT_PAL_M;
			break;
		case 4:
			buf3.arg1 = DXR2_OUTPUTFORMAT_PAL_M_60;
			break;
		case 5:
			buf3.arg1 = DXR2_OUTPUTFORMAT_PAL_N;
			break;
		case 6:
			buf3.arg1 = DXR2_OUTPUTFORMAT_PAL_Nc;
			break;
		case 7:
			buf3.arg1 = DXR2_OUTPUTFORMAT_PAL_60;
			break;
		default:
			return -1;
		}

		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_TV_OUTPUT_FORMAT, &buf3))
			return -1;
	}

	if (decoder.status == DXR2_PLAYMODE_PLAY && attr->fps != -1)
		ioctl(decoder.fd_dxr2,DXR2_IOC_PLAY,NULL);

	return 0;
}


/**
 *
 */

int DecoderSetAudioAttributes (plugin_codec_attr_audio_t *attr)
{
	dxr2_twoArg_t buf3;

	if (attr->volume != -1) {
		if (attr->volume > 100 || attr->volume < 0)
			return -1;

		buf3.arg1 = attr->volume*19/100;

		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_AUDIO_VOLUME, &buf3))
			return -1;
	}

	if (attr->freq != -1) {
		switch (attr->freq) {
		case 0:	
			buf3.arg1 = DXR2_AUDIO_FREQ_441;
			break;
		case 1:
			buf3.arg1 = DXR2_AUDIO_FREQ_48;
			break;
		case 2:
			buf3.arg1 = DXR2_AUDIO_FREQ_96;
			break;
		case 3:
			buf3.arg1 = DXR2_AUDIO_FREQ_2205;
			break;
		case 4:
			buf3.arg1 = DXR2_AUDIO_FREQ_32;
			break;
		default:
			return -1;
		}

		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_AUDIO_SAMPLE_FREQUENCY, &buf3))
			return -1;
	}

	if (attr->bits != -1) {
		switch (attr->bits) {
		case 0:
			buf3.arg1 = DXR2_AUDIO_WIDTH_16;
			break;
		case 1:
			buf3.arg1 = DXR2_AUDIO_WIDTH_20;
			break;
		case 2:
			buf3.arg1 = DXR2_AUDIO_WIDTH_24;
			break;
		default:
			return -1;
		}

		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SET_AUDIO_DATA_WIDTH, &buf3))
			return -1;
	}

	if (attr->mode != -1) {
		switch (attr->mode) {
		case 0:
			decoder.current_audio_mode = DXR2_STREAM_AUDIO_AC3;
			break;
		case 1:
			decoder.current_audio_mode = DXR2_STREAM_AUDIO_MPEG;
			break;
		case 2:
			decoder.current_audio_mode = DXR2_STREAM_AUDIO_LPCM;
			break;
		default:
			return -1;
		}
	}

	if (attr->stream != -1)
		decoder.current_audio_stream = attr->stream;

	if (attr->stream != -1 || attr->mode != -1) {
		if (decoder.current_audio_mode==-1)
			return -1;

		buf3.arg1 = decoder.current_audio_mode;
		buf3.arg2 = decoder.current_audio_stream;

		if (ioctl(decoder.fd_dxr2, DXR2_IOC_SELECT_STREAM, &buf3))
			return -1;
	}

	return 0;
}


/**
 *
 */

int DecoderSPU (int n) {
	dxr2_threeArg_t buf3;

	if (n == -1) {
		buf3.arg1 = DXR2_SUBPICTURE_OFF;
		return ioctl(decoder.fd_dxr2, DXR2_IOC_ENABLE_SUBPICTURE, &buf3);
	} else {
		buf3.arg1 = DXR2_SUBPICTURE_ON;
		if (ioctl(decoder.fd_dxr2, DXR2_IOC_ENABLE_SUBPICTURE, &buf3))
			return -1;

		buf3.arg1 = DXR2_STREAM_SUBPICTURE;
		buf3.arg2 = n;
		return ioctl(decoder.fd_dxr2, DXR2_IOC_SELECT_STREAM, &buf3);
	}
}


/**
 *
 */

int DecoderSetAttributes (u_int id, void *attr)
{
	int ret = 0;
	pthread_mutex_lock(&(decoder.mutex));

	switch (id) {
	case BUF_AUDIO: 
		if (DecoderSetAudioAttributes ((plugin_codec_attr_audio_t *) attr))
			ret = -1;
		break;
	case BUF_VIDEO:
		if (DecoderSetVideoAttributes ((plugin_codec_attr_video_t *) attr))
			ret = -1;
		break;
	case BUF_SUBPIC: {
		plugin_codec_attr_spu_t *spu = (plugin_codec_attr_spu_t *) attr;

		if (spu->clut != NULL)
			 if (DecoderSPUCLUT (spu->clut))
				ret = -1;

		if (DecoderSPU (spu->num))
			ret = -1;
		break;
	}
	}

	pthread_mutex_unlock(&(decoder.mutex));

	return ret;
}


/**
 *
 */

int DecoderSPUCLUT(char *data)
{
	char clut[64];
	int i;

	for (i=0; i<16; i++) {
		clut[i*4+0] = data[i*4+3];
		clut[i*4+1] = data[i*4+2];
		clut[i*4+2] = data[i*4+1];
		clut[i*4+3] = data[i*4+0];
	}

	return ioctl (decoder.fd_dxr2, DXR2_IOC_SET_SUBPICTURE_PALETTE, clut);
}


/**
 *
 */

int DecoderSetStatus (u_int status)
{
	switch (status) {
	case STATUS_PLAY:
		fprintf(stderr,"DecoderPlay\n");

		if (decoder.status == DXR2_PLAYMODE_PAUSED) {
			pthread_mutex_lock(&(decoder.mutex));
			pthread_cond_signal (&(decoder.playing));
			pthread_mutex_unlock(&(decoder.mutex));
			return 0;
		}
		decoder.end_loop = 0;
		pthread_create (&decoder.loop, NULL, _DecoderLoop, NULL);
		break;
	case STATUS_STOP:
		fprintf(stderr,"DecoderStop\n");

		decoder.end_loop = 1;
		pthread_join (decoder.loop, NULL);

		if (decoder.status == DXR2_PLAYMODE_PAUSED) {
			pthread_mutex_lock(&(decoder.mutex));
			pthread_cond_signal (&(decoder.playing));
			pthread_mutex_unlock(&(decoder.mutex));
		}


		if (ioctl(decoder.fd_dxr2,DXR2_IOC_STOP,NULL))
			return -1;

		decoder.status=DXR2_PLAYMODE_STOPPED;

		return ioctl(decoder.fd_dxr2,DXR2_IOC_CLEAR_VIDEO,NULL);
		break;
	case STATUS_PAUSE:
		fprintf(stderr,"DecoderPause\n");

		decoder.status = DXR2_PLAYMODE_PAUSED;
		break;
	default:
		fprintf (stderr, "DecoderSetStatus: unhandled status: %d\n", status);
		return -1;
	}

	return 0;
}

#if 0	// sorry MATS, but due to the stat of putting this into a shared
	// object, i try to keep things simple (for the moment - will 
	// be merged in later

int Decoder_CSS_Disc_Key(char *data) {
  return ioctl(decoder.fd_dxr2, DXR2_IOC_SEND_DISC_KEY, data);
}

int Decoder_CSS_Title_Key(char *data, int cpm, int cp_sec, int cgms) {
  dxr2_oneArg_t argBuf;
  dxr2_titleKey_t key;

  key.cgmsFlags=(cpm << 7) | (cp_sec << 6) | (cgms << 4);
  memcpy(key.key,data,5);

  if (ioctl(decoder.fd_dxr2, DXR2_IOC_SEND_TITLE_KEY, &key)) return -1;

  argBuf.arg = DXR2_CSSDECRMODE_ON;
  return ioctl(decoder.fd_dxr2, DXR2_IOC_SET_DECRYPTION_MODE, &argBuf);
}

int Decoder_CSS_Drive_Auth_Challenge(char *data) {
  return ioctl(decoder.fd_dxr2, DXR2_IOC_GET_CHALLENGE_KEY, data);
}
int Decoder_CSS_Drive_Auth_Response(char *data) {
  return ioctl(decoder.fd_dxr2, DXR2_IOC_SEND_RESPONSE_KEY, data);
}

int Decoder_CSS_Host_Auth(char *data) {
  if (ioctl(decoder.fd_dxr2, DXR2_IOC_SEND_CHALLENGE_KEY, data)) return -1;
  return ioctl(decoder.fd_dxr2, DXR2_IOC_GET_RESPONSE_KEY, data);
}
#endif
